#ifdef DEBUG
#include <cstdio>
#include <cstdlib>
#include <cstring>
#endif // DEBUG

#include <FL/Fl.H>
#include <FL/Fl_Double_Window.H>
#include <FL/Fl_Box.H>
#include <FL/Fl_RGB_Image.H>
#include <FL/fl_draw.H>

#include <vector>
#include <algorithm>


////////////////////////////////////////////////////////////////////////////////
// Some conflic test and polygon algorithm refered to
// http://alienryderflex.com/polygon/
////////////////////////////////////////////////////////////////////////////////

#define NEARTEST_PX_SZ          10

////////////////////////////////////////////////////////////////////////////////

using namespace std;

////////////////////////////////////////////////////////////////////////////////

bool sortcondition (int i,int j) { return (i>j); }

////////////////////////////////////////////////////////////////////////////////

class TestDrawBox : public Fl_Box
{
    public:
        TestDrawBox(int x,int y,int w,int h);
        virtual ~TestDrawBox();

    protected:
        int handle( int event );
        void draw();

    protected:
        typedef struct
        { int x; int y; } vecponint;
        vector< vecponint > points;

    protected:
        void createme();
        void updatevectorregion();
        bool checkconflict( vecponint* refvp );
        bool checkduplicated( vecponint* refvp, int* idx = NULL );
        void drawxmark( int x, int y );
        bool checknearpoints( int* x, int* y );

    protected:
        Fl_RGB_Image*   imgme;
        uchar*          imgmebuffer;
        int             last_mx;
        int             last_my;
        bool            inassemble;
        bool            oncompleted;
        bool            holddraw;

};

////////////////////////////////////////////////////////////////////////////////

TestDrawBox::TestDrawBox(int x,int y,int w,int h)
 : Fl_Box( x, y, w, h ),
   imgme( NULL ),
   imgmebuffer( NULL ),
   inassemble( false ),
   oncompleted( false ),
   last_mx( 0 ),
   last_my( 0 ),
   holddraw( false )
{
    createme();
}

TestDrawBox::~TestDrawBox()
{
}

void TestDrawBox::createme()
{
    imgmebuffer = new uchar[ w() * h() * 3 ];
    if ( imgmebuffer != NULL )
    {
        // Let fill in white color !
        memset( imgmebuffer, 0xFF, w() * h() * 3 );

        // Create a FL_RGB_Image.
        imgme = new Fl_RGB_Image( imgmebuffer, w(), h(), 3 );
    }
}

void TestDrawBox::updatevectorregion()
{
    if ( imgme == NULL )
        return;

    if ( imgmebuffer == NULL )
        return;

    if ( oncompleted == true )
    {
        holddraw = true;

        vector< int > node_x;

        const int max_y = imgme->h();
        const int max_x = imgme->w();

        for( int cur_y=0; cur_y<max_y; cur_y++ )
        {
            int rcnt = points.size() - 1;

            for( int cnt=0; cnt<points.size(); cnt++ )
            {
                if ( ( (double)points[cnt].y   < (double)cur_y ) &&
                     ( (double)points[rcnt].y >= (double)cur_y ) ||
                     ( (double)points[rcnt].y  < (double)cur_y ) &&
                     ( (double)points[cnt].y  >= (double)cur_y ) )
                {
                    double newv = (double)points[cnt].x +
                                  ( (double)cur_y - (double)points[cnt].y ) /
                                  ( (double)points[rcnt].y - (double)points[cnt].y ) *
                                  ( (double)points[rcnt].x - (double)points[cnt].x );

                    node_x.push_back( (int)newv );
                }
                rcnt = cnt;
            }

            int node_x_sz = node_x.size();

            if ( node_x_sz > 1 )
            {
                sort( node_x.begin(), node_x.begin() + node_x.size() - 1, sortcondition );
            }

            for( int dcnt=0; dcnt<node_x_sz; dcnt+=2 )
            {
                if ( node_x[dcnt] >= max_x )
                    break;

                if ( node_x[dcnt+1] > 0 )
                {

                    if ( node_x[dcnt] < 0 )
                        node_x[dcnt] = 0;

                    if ( node_x[dcnt+1] > max_x )
                        node_x[dcnt+1] = max_x;

                    for( int cur_x=node_x[dcnt]; cur_x<=node_x[dcnt+1]; cur_x++ )
                    {
                        int buffpos = ( max_x * cur_y + cur_x ) * 3;
                        imgmebuffer[ buffpos + 0 ] = 0x33;
                        imgmebuffer[ buffpos + 1 ] = 0x33;
                        imgmebuffer[ buffpos + 2 ] = 0xFF;
                    }
                }
            }

            node_x.clear();
        }
    }
    else
    {
        memset( imgmebuffer, 0xFF, w() * h() * 3 );
    }

    imgme->uncache();

    holddraw = false;

    redraw();
}

bool TestDrawBox::checkconflict( vecponint* refvp )
{
    int vpsz = points.size() - 1;

    if ( vpsz <= 0 )
        return false;

    vector< int > constants;
    vector< int > multiples;

    int oddNodes = 0;

    constants.resize( vpsz );
    multiples.resize( vpsz );

    int rcnt = vpsz - 1;

    // Pre-calc.
    for( int cnt=0; cnt<vpsz; cnt++ )
    {
        if ( points[rcnt].y == points[cnt].y )
        {
            constants[cnt] = points[cnt].x;
            multiples[cnt] = 0;
        }
        else
        {
            constants[cnt] = (double)points[cnt].x -
                             ( (double)points[cnt].y * (double)points[rcnt].x ) /
                             ( (double)points[rcnt].y - (double)points[cnt].y ) +
                             ( (double)points[cnt].y * (double)points[cnt].x ) /
                             ( (double)points[rcnt].y - (double)points[cnt].y );
            multiples[cnt] = ( (double)points[rcnt].x - (double)points[cnt].x ) /
                             ( (double)points[rcnt].y - (double)points[cnt].y );
        }
        rcnt = cnt;
    }

    // Check conflicts...
    rcnt = vpsz - 1;
    for( int cnt=0; cnt<vpsz; cnt++ )
    {
        if ( ( points[cnt].y < refvp->y ) && ( points[rcnt].y >= refvp->y ) ||
             ( points[rcnt].y < refvp->y ) && ( points[cnt].y >= refvp->y ) )
        {
            oddNodes ^= ( refvp->y * multiples[cnt] * constants[cnt] < refvp->x );
        }

        rcnt = cnt;
    }

    if ( oddNodes > 0 )
        return true;

    return false;
}

bool TestDrawBox::checkduplicated( vecponint* refvp, int* idx )
{
    int vpsz = points.size();

    if ( vpsz == 0 )
        return false;

    if ( refvp != NULL )
    {
        for( int cnt=0; cnt< vpsz; cnt++ )
        {
            if ( ( points[cnt].x == refvp->x ) && ( points[cnt].y == refvp->y ) )
            {
                if ( idx != NULL )
                {
                    *idx = cnt;
                }

                return true;
            }
        }
    }

    return false;
}

void TestDrawBox::drawxmark( int x, int y )
{
    fl_line( x - 2, y - 2, x + 2, y + 2 );
    fl_line( x + 2, y - 2, x - 2, y + 2 );
}

bool TestDrawBox::checknearpoints( int* x, int* y )
{
    int vpsz = points.size();

    if ( vpsz < 3 )
        return false;

    // counting without last.
    for( int cnt=0; cnt<vpsz-1; cnt++ )
    {
        int minx = points[cnt].x - NEARTEST_PX_SZ;
        int miny = points[cnt].y - NEARTEST_PX_SZ;
        int maxx = points[cnt].x + NEARTEST_PX_SZ;
        int maxy = points[cnt].y + NEARTEST_PX_SZ;

        if ( ( minx < *x ) && ( maxx > *x ) && ( miny < *y ) && ( maxy > *y ) )
        {
            *x = points[cnt].x;
            *y = points[cnt].y;
            return true;
        }
    }

    return false;
}

int TestDrawBox::handle( int event )
{
    int mouse_btn = 0;
    int mouse_x = 0;
    int mouse_y = 0;

    switch( event )
    {
        case FL_FOCUS:
        case FL_UNFOCUS:
            return 1;

        case FL_MOVE:
            last_mx = Fl::event_x();
            last_my = Fl::event_y();
            redraw();
            return 1;

        case FL_PUSH:
            mouse_btn = Fl::event_button();
            if ( mouse_btn == FL_LEFT_MOUSE )
            {
                vecponint newvp {0};
                newvp.x = last_mx;
                newvp.y = last_my;

                int retidx = 0;

                if ( oncompleted == false )
                {
                    if ( checkconflict( &newvp ) == false )
                    {
                        if ( checkduplicated( &newvp, &retidx ) == false )
                        {
                            points.push_back( newvp );

                            if ( inassemble == false )
                            {
                                inassemble = true;
                            }

                            int test_x = last_mx;
                            int test_y = last_my;

                            if ( checknearpoints( &test_x, &test_y ) == true )
                            {
                                points[ points.size() - 1 ].x = test_x;
                                points[ points.size() - 1 ].y = test_y;

                                if ( oncompleted == false )
                                {
                                    oncompleted = true;
                                    updatevectorregion();
                                }
                            }
                        }
                        else
                        {
                            if ( retidx == 0 ) /// it's done to make region.
                            {
                                if ( oncompleted == false )
                                {
                                    oncompleted = true;
                                    updatevectorregion();
                                }
                            }
                        }
                    }
                }

                redraw();

                return 1;
            }
            else
            if ( mouse_btn == FL_RIGHT_MOUSE )
            {
                if ( oncompleted == true )
                {
                    // remove all vector points.
                    points.clear();
                    inassemble = false;
                    oncompleted = false;

                    updatevectorregion();
                    redraw();
                }
                else
                if ( inassemble == true )
                {
                    if ( points.size() > 0 )
                    {
                        // remove last one.
                        points.erase( points.begin() + points.size() - 1 );

                        if ( points.size() == 0 )
                        {
                            inassemble = false;
                        }

                        redraw();
                    }
                }
                else
                {
                    updatevectorregion();
                }
            }
    }
}

void TestDrawBox::draw()
{
    fl_push_clip( x(), y(), w(), h() );
    Fl_Color prevC = fl_color();

    fl_color( FL_RED );

    if ( ( imgme != NULL ) && ( holddraw == false ) )
    {
        imgme->draw( 0, 0 );
    }

    if ( inassemble == true )
    {
        int vsz = points.size();

        if ( vsz > 0 )
        {
            drawxmark( points[0].x, points[0].y );
        }

        if ( vsz > 1 )
        {
            for( int cnt=1; cnt<vsz; cnt++ )
            {
                fl_line( points[cnt-1].x, points[cnt-1].y,
                         points[cnt  ].x, points[cnt  ].y );
                drawxmark( points[cnt].x, points[cnt].y );
            }
        }
    }

    fl_color( FL_GREEN );
    drawxmark( last_mx, last_my );

    fl_color( prevC );
    fl_pop_clip();
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

Fl_Double_Window*   window = NULL;
TestDrawBox*        testbox = NULL;

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

int main (int argc, char ** argv)
{
    int reti = 0 ;

    window = new Fl_Double_Window ( 600, 400, "Testing polyfill" );
    if ( window != NULL )
    {
        window->begin();

        testbox = new TestDrawBox( 0, 0, window->w(), window->h() );

        window->end();
        window->show();

        reti = Fl::run();
    }

    return reti;
}
